/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Movement
Version:   1.0
ID:        ID_MOVEMENT
Category:  GUI
Date:      February 2003
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2000-2003.  All rights reserved.
Keywords:  movement, move, mouse, pointer, motion
Short:     Used to react to the user's movement within the GUI.

###DESCRIPTION###
<p>The Movement class is used to detect the movement of any 'pointing
device' that the user is moving on the display.  In the vast majority of
cases this means dealing with mouse pointers, although touch screens can
also express movement, as well as joysticks and other devices that can
move the pointer.</p>

<p>The Movement class supports a number of native actions that can be
performed when movement is detected, including:</p>

<dir>
  <li>Can move objects to the back or front of the display.</li>
  <li>Can alter the frame of its monitored object.</li>
  <li>Will send Activate messages to all children of the movement object,
  which can then perform whatever commands you specified.</li>
</dir>

<p>By default, a movement object will monitor the user's motion in the local
container.  The handling of movement is split into
three parts - movement entry, internal movement and movement exit.  The
default behaviour is to monitor movement entry, which causes the movement
object to react only when the pointer moves over the object for the first
time.  You can do the opposite and monitor for a user exit by setting the
Exit field to TRUE.</p>

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/graphics/render.h>
#include <pandora/graphics/pointer.h>
#include <pandora/modules/pointer.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_MOVEMENT  1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_MOVEMENT, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2004.  All rights reserved.",
   "January 2004",
   "Movement"
};

static struct KernelBase *KernelBase;
static struct PointerBase *PointerBase;
static OBJECTPTR modPointer = NULL;
static OBJECTPTR MovementClass = 0;

/*****************************************************************************
** Movement class definition.
*/

#define LEN_POINTER 28

struct Movement {
   OBJECT_HEADER          /* Class header */
   OBJECTID MoveToFront;  /* Object to move to front */
   OBJECTID MoveToBack;   /* Object to move to back */
   OBJECTID DrawableID;   /* Refers to the drawable that is being monitored for movement */
   LONG     Frame;        /* The frame to change the Render object to when movement occurs */
   LONG     ExitFrame;    /* Frame to change to when the user leaves the Render object */
   LONG     Exit;         /* Perform the action when the user leaves the Render object */
   LONG     Enter;        /* Perform the action when the user enters the Render object */
   LONG     State;        /* Reflects the current state of the Movement object */

   PRIVATE_FIELDS
   FLOAT  XCoord;
   FLOAT  YCoord;
   FLOAT  Width;
   FLOAT  Height;
   FLOAT  YOffset;
   FLOAT  XOffset;
   LONG   Dimensions;
   UBYTE  Pointer[LEN_POINTER];
};

#define STATE_ENTERED  1
#define STATE_INSIDE   2
#define STATE_EXITED   3

static ERROR GET_Height(struct Movement *, struct Variable *);
static ERROR GET_Pointer(struct Movement *, STRING *);
static ERROR GET_Width(struct Movement *, struct Variable *);
static ERROR GET_X(struct Movement *, struct Variable *);
static ERROR GET_XOffset(struct Movement *, struct Variable *);
static ERROR GET_Y(struct Movement *, struct Variable *);
static ERROR GET_YOffset(struct Movement *, struct Variable *);

static ERROR SET_Enter(struct Movement *, LONG *);
static ERROR SET_Exit(struct Movement *, LONG *);
static ERROR SET_Height(struct Movement *, struct Variable *);
static ERROR SET_Pointer(struct Movement *, STRING);
static ERROR SET_Width(struct Movement *, struct Variable *);
static ERROR SET_X(struct Movement *, struct Variable *);
static ERROR SET_XOffset(struct Movement *, struct Variable *);
static ERROR SET_Y(struct Movement *, struct Variable *);
static ERROR SET_YOffset(struct Movement *, struct Variable *);

static struct FieldArray MovementFields[] = {
   { "MoveToFront", 0, FDF_OBJECTID|FDF_RW, 0, NULL, NULL },
   { "MoveToBack",  0, FDF_OBJECTID|FDF_RW, 0, NULL, NULL },
   { "Drawable",    0, FDF_OBJECTID|FDF_RI, ID_RENDER, NULL, NULL },
   { "Frame",       0, FDF_LONG|FDF_RW,     0, NULL, NULL },
   { "ExitFrame",   0, FDF_LONG|FDF_RW,     0, NULL, NULL },
   { "Exit",        0, FDF_LONG|FDF_RW,     0, NULL, SET_Exit },
   { "Enter",       0, FDF_LONG|FDF_RW,     0, NULL, SET_Enter },
   { "State",       0, FDF_LONG|FDF_R,      0, NULL, NULL },
   /*** Virtual fields ***/
   { "Height",      0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Height,  SET_Height },
   { "Pointer",     0, FDF_STRING|FDF_RW,                             0, GET_Pointer, SET_Pointer },
   { "Width",       0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Width,   SET_Width },
   { "X",           0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XCoord",      0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XOffset",     0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_XOffset, SET_XOffset },
   { "Y",           0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YCoord",      0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YOffset",     0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_YOffset, SET_YOffset },
   END_FIELD
};

static ERROR MOVE_ActionNotify(struct Movement *, struct acActionNotify *);
static ERROR MOVE_ClosingTag(struct Movement *, APTR);
static ERROR MOVE_Init(struct Movement *, APTR);

static struct ActionArray MovementActions[] = {
   { AC_ActionNotify, MOVE_ActionNotify },
   { AC_ClosingTag,   MOVE_ClosingTag },
   { AC_Init,         MOVE_Init },
   { NULL, NULL }
};

/*****************************************************************************
** Command: Init()
*/

ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &modPointer, NULL,
      FID_Name|TSTRING,   "pointer",
      FID_Version|TFLOAT, MODVERSION_POINTER,
      TAGEND) IS ERR_Okay) {
      if (GetField(modPointer, FID_ModBase, FT_POINTER, &PointerBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&MovementClass, NULL,
      FID_BaseClassID|TLONG, ID_MOVEMENT,
      FID_Name|TSTRING,      "Movement",
      FID_Category|TLONG,    CCF_GUI,
      FID_Actions|TPTR,      MovementActions,
      FID_Fields|TPTR,       MovementFields,
      FID_Size|TLONG,        sizeof(struct Movement),
      FID_Version|TFLOAT,    VER_MOVEMENT,
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (modPointer)    { acFree(modPointer); modPointer = NULL; }
   if (MovementClass) { acFree(MovementClass); MovementClass = 0; }
   return(ERR_Okay);
}

/*****************************************************************************
** Movement: ActionNotify
*/

static ERROR MOVE_ActionNotify(struct Movement *Self, struct acActionNotify *NotifyArgs)
{
   struct acUserMovement *Args;
   struct ChildEntry list[30];
   OBJECTPTR Drawable;
   LONG width, height, xcoord, ycoord, i, count;
   UBYTE inside;

   if ((NotifyArgs->ActionID IS AC_UserMovement) AND (NotifyArgs->Error IS ERR_Okay)) {
      /*** Determine the current movement state (exit, enter, inside) ***/

      Args = (struct acUserMovement *)NotifyArgs->Args;

      if (Self->Dimensions) {
         if (GetFields(Self, FID_Width|TLONG,  &width,
                             FID_Height|TLONG, &height,
                             FID_XCoord|TLONG, &xcoord,
                             FID_YCoord|TLONG, &ycoord,
                             TAGEND) != ERR_Okay) return(ERR_GetField);
      }


      inside = TRUE;
      if (Args->OverObject IS Self->DrawableID) {
         if (Self->Dimensions) {
            if ((Args->XCoord < xcoord) OR (Args->YCoord < ycoord) OR
                (Args->XCoord >= xcoord + width) OR (Args->YCoord >= ycoord + height)) {
               inside = FALSE;
            }
         }
      }
      else inside = FALSE;

      if (inside) {

         if (Self->Pointer[0]) {
            SetCursor(NULL, ACF_BUFFER, 0, Self->Pointer, Self->Head.UniqueID, NULL);
         }

         if (Self->State IS STATE_ENTERED) Self->State = STATE_INSIDE;
         else if (Self->State != STATE_INSIDE) {
            Self->State = STATE_ENTERED;
         }
      }
      else {
         if (Self->State IS STATE_EXITED) return(ERR_Okay);
         else {
            Self->State = STATE_EXITED;

            if (Self->Pointer[0]) {
               RestoreCursor(PTR_DEFAULT, Self->Head.UniqueID);
            }
         }
      }

      /*** Change the drawable's frame if necessary ***/

      if (Self->DrawableID) {
         if (AccessObject(Self->DrawableID, 2000, (OBJECTPTR *)&Drawable) IS ERR_Okay) {
            if (Self->State IS STATE_EXITED) {
               SetField(Drawable, FID_Frame, FT_LONG, &Self->ExitFrame);
               if (Self->Dimensions) ActionTags(AC_Draw, Drawable, xcoord, ycoord, width, height);
               else Action(AC_Draw, Drawable, NULL);
            }
            else if (Self->State IS STATE_ENTERED) {
               SetField(Drawable, FID_Frame, FT_LONG, &Self->Frame);
               if (Self->Dimensions) ActionTags(AC_Draw, Drawable, xcoord, ycoord, width, height);
               else Action(AC_Draw, Drawable, NULL);

               /*** Move to front and move to back options ***/

               if (Self->MoveToFront) {
                  if (ActionMsg(AC_MoveToFront, Self->MoveToFront, NULL) IS ERR_NoMatchingObject) {
                     Self->MoveToFront = NULL;
                  }
               }
               else if (Self->MoveToBack) {
                  if (ActionMsg(AC_MoveToBack, Self->MoveToBack, NULL) IS ERR_NoMatchingObject) {
                     Self->MoveToBack = NULL;
                  }
               }
            }
            ReleaseObject(Drawable);
         }
      }

      if ((Self->Exit IS TRUE) AND (Self->State != STATE_EXITED)) return(ERR_Okay);
      else if ((Self->Enter IS TRUE) AND (Self->State != STATE_ENTERED)) return(ERR_Okay);

      /* This class has a responsibility to activate any
      ** children that may be attached to the Movement object.
      */

      if ((Self->State IS STATE_ENTERED) AND (Self->Enter)) {
         count = ARRAYSIZE(list);
         if (ListChildren(GetUniqueID(Self), list, &count) IS ERR_Okay) {
            for (i=0; i < count; i++) ActionMsg(AC_Activate, list[i].ObjectID, NULL);
         }
      }
      else if ((Self->State IS STATE_EXITED) AND (Self->Exit)) {
         count = ARRAYSIZE(list);
         if (ListChildren(GetUniqueID(Self), list, &count) IS ERR_Okay) {
            for (i=0; i < count; i++) ActionMsg(AC_Activate, list[i].ObjectID, NULL);
         }
      }

      /*** Record change of state ***/

      if (Self->State IS STATE_EXITED); // Self->State = NULL;
      else if (Self->State IS STATE_ENTERED) Self->State = STATE_INSIDE;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Movement: ClosingTag
*/

static ERROR MOVE_ClosingTag(struct Movement *Self, APTR Args)
{
   /*** Check if we are associated with a Drawable object (this is optional) ***/

   if (!Self->DrawableID) {
      OBJECTID OwnerID;
      OwnerID = GetOwner(Self);
      while ((OwnerID) AND (GetClassID(OwnerID) != ID_RENDER)) {
         OwnerID = GetOwnerID(OwnerID);
      }
      if (OwnerID) Self->DrawableID = OwnerID;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Movement: Init
*/

static ERROR MOVE_Init(struct Movement *Self, APTR Void)
{
   OBJECTPTR owner;
   if (AccessObject(GetOwner(Self), 2000, &owner) IS ERR_Okay) {
      SubscribeAction(owner, AC_UserMovement, Self);
      ReleaseObject(owner);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Drawable
Short:  Refers to the drawable to be used for movement monitoring.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>In most cases, when you create a Movement object it should be
contained by a drawable area, such as a Render object.  However there may
be situations where you want to initialise a Movement object to a
non-graphical container.  If this is the case then you must set the
Drawable field so that the Movement object knows what drawable it needs to
monitor for movement.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Enter
Short:  Set to TRUE to activate child object/s when the user enters the monitored area.
Type:   BOOLEAN
Status: Read/Write

###DESCRIPTION###
<p>If this field is set to TRUE, the Movement object will activate all
of its child objects when the user enters the monitored area.</p>

###SEE ALSO###
Field: Exit

###END###

*****************************************************************************/

static ERROR SET_Enter(struct Movement *Self, LONG *Value)
{
   if (*Value IS TRUE) {
      Self->Enter = TRUE;
      Self->Exit  = FALSE;
   }
   else Self->Enter = FALSE;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Exit
Short:  Set to TRUE to activate child object/s when the user leaves the monitored area.
Type:   BOOLEAN
Status: Read/Write

###DESCRIPTION###
<p>If this field is set to TRUE, the Movement object will activate all
of its child objects when the user leaves the monitored area.</p>

###SEE ALSO###
Field: Enter

###END###

*****************************************************************************/

static ERROR SET_Exit(struct Movement *Self, LONG *Value)
{
   if (*Value IS TRUE) {
      Self->Exit  = TRUE;
      Self->Enter = FALSE;
   }
   else Self->Exit = FALSE;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ExitFrame
Short:  Enables frame switching when the user leaves the monitored object.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>If you would like the Movement object to change the frame number of its
related drawable when the user leaves the object, set this field to the frame
number that you want to use.</p>

###SEE ALSO###
Fields: Drawable, Frame

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Frame
Short:  Enables frame switching when the user enters the monitored object.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>If you would like the Movement object to change the frame number of its
related drawable when the user enters the object, set this field to the
frame  number that you want to use.</p>

###SEE ALSO###
Fields: Drawable, ExitFrame

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Height
Short:  Defines the height of a Movement object's monitored area.
Type:   FLOAT
Status: Read/Write

###DESCRIPTION###
<p>If you want a Movement object to monitor a specific region of the drawable
that it is connected to, set the XCoord, YCoord, Width and Height fields.  If
you would rather have the Movement object monitor the entire area of the
drawable, do not set any of the fields.</p>

<p>To set a relative height, use the FD_PERCENT flag when setting the field;
otherwise a fixed height is assumed.</p>

###SEE ALSO###
Field: Width

###END###

*****************************************************************************/

static ERROR GET_Height(struct Movement *Self, struct Variable *Value)
{
   OBJECTPTR Render;
   FLOAT ContainerHeight, YCoord, YOffset;

   if (Value->Type IS FD_DOUBLE) {
      if (Self->Dimensions & DMF_FIXED_HEIGHT) {
         Value->Double = Self->Height;
      }
      else if (Self->Dimensions & DMF_RELATIVE_HEIGHT) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Height, FT_FLOAT, &ContainerHeight);
            Value->Double = Self->Height * ContainerHeight / 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Height, FD_FLOAT, &ContainerHeight);
            GetFields(Self, FID_YCoord|TFLOAT,  &YCoord,
                            FID_YOffset|TFLOAT, &YOffset,
                            TAGEND);
            Value->Double = ContainerHeight - YOffset - YCoord;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else Value->Double = 0;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Value->Double = 0;
      if (Self->Dimensions & DMF_RELATIVE_HEIGHT) {
         Value->Double = Self->Height;
      }
      else if (Self->Dimensions & DMF_FIXED_HEIGHT) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Height, FT_FLOAT, &ContainerHeight);
            Value->Double = Self->Height / ContainerHeight * 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Height, FD_FLOAT, &ContainerHeight);
            if (Self->Dimensions & DMF_FIXED_Y) YCoord = Self->YCoord;
            else YCoord = Self->YCoord * 100 / ContainerHeight;
            if (Self->Dimensions & DMF_FIXED_YOFFSET) YOffset = Self->YOffset;
            else YOffset = Self->YOffset * 100 / ContainerHeight;
            Value->Double = (ContainerHeight - YOffset - YCoord) / ContainerHeight * 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Height(struct Movement *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      if (Value->Double <= 0) {
         DPrintF("!SetField:","[Movement] A height setting of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_HEIGHT) | DMF_FIXED_HEIGHT;
      Self->Height     = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      if (Value->Double <= 0) {
         DPrintF("!SetField:","[Movement] A height setting of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_HEIGHT) | DMF_RELATIVE_HEIGHT;
      Self->Height     = Value->Double;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   MoveToBack
Short:  Can refer to an object that should be moved to the back when movement is detected.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>When the user enters the monitored area, you can move a selected object
to the back of its container by setting this field to that object's ID.
Here is an example that moves a Render object to the back when the user
enters it:</p>

<pre>
  &lt;render name="drawable" width="100" height="100" colour="#606000"&gt;
    &lt;movement object="drawable" movetoback="[drawable]"/&gt;
  &lt;/render&gt;
</pre>


###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   MoveToFront
Short:  Can refer to an object that should be moved to the front when movement is detected.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>When the user enters the monitored area, you can move a selected object
to the front of its container by setting this field to that object's ID.
Here is an example that moves a Render object to the front when the user
enters it:</p>

<pre>
  &lt;render name="drawable" width="100" height="100" colour="#606000"&gt;
    &lt;movement object="drawable" movetofront="[drawable]"/&gt;
  &lt;/render&gt;
</pre>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   State
Short:  Reflects the current state of the movement object.
Type:   LONG
Status: Read

###DESCRIPTION###
<p>The current state of the Movement object in relation to the user is
always reflected in this field.  The numbers that may be returned from the
field are as follows:</p>

<pre>
  ENTERED = 1
  INSIDE  = 2
  EXITED  = 3
</pre>

<p>The ENTERED state means that the user has just entered the monitored area.
The INSIDE state means that the user is moving within the monitored area.
The EXITED state means that the user has left the monitored area.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Pointer
Short:  Defines the pointer cursor to use when the pointer enters the vicinity of the movement area.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>Set the Pointer field if you would like to use a special pointer image
when the user's pointer enters the movement area.  Valid image names for the
pointer are defined in the Pointer class documentation.</p>

###SEE ALSO###
Class: Pointer

###END###

*****************************************************************************/

static ERROR GET_Pointer(struct Movement *Self, STRING *Value)
{
   *Value = Self->Pointer;
   return(ERR_Okay);
}

static ERROR SET_Pointer(struct Movement *Self, STRING Value)
{
   WORD i;
   for (i=0; (i < LEN_POINTER-1) AND (Value[i]); i++) Self->Pointer[i] = Value[i];
   Self->Pointer[i] = 0;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Width
Short:  Defines the width of a Movement object's monitored area.
Type:   FLOAT
Status: Get/Set

###DESCRIPTION###
<p>If you want a Movement object to monitor a specific region of the drawable
that it is connected to, set the XCoord, YCoord, Width and Height fields.  If
you would rather have the Movement object monitor the entire area of the
drawable, do not set any of the fields.</p>

<p>To set a relative width, use the FD_PERCENT flag when setting the field;
otherwise a fixed width is assumed.</p>

###SEE ALSO###
Field: Height

###END###

*****************************************************************************/

static ERROR GET_Width(struct Movement *Self, struct Variable *Value)
{
   OBJECTPTR Render;
   FLOAT ContainerWidth, XCoord, XOffset;

   if (Value->Type IS FD_DOUBLE) {
      Value->Double = 0;
      if (Self->Dimensions & DMF_FIXED_WIDTH) {
         Value->Double = Self->Width;
      }
      else if (Self->Dimensions & DMF_RELATIVE_WIDTH) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Width, FT_FLOAT, &ContainerWidth);
            Value->Double = Self->Width * ContainerWidth / 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Width, FT_FLOAT, &ContainerWidth);
            if (Self->Dimensions & DMF_FIXED_X) XCoord = Self->XCoord;
            else XCoord = ContainerWidth * Self->XCoord / 100;
            if (Self->Dimensions & DMF_FIXED_XOFFSET) Value->Double = ContainerWidth - XCoord - Self->XOffset;
            else Value->Double = ContainerWidth - XCoord - (ContainerWidth * Self->XOffset / 100);
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Value->Double = 0;
      if (Self->Dimensions & DMF_RELATIVE_WIDTH) {
         Value->Double = Self->Width;
      }
      else if (Self->Dimensions & DMF_FIXED_WIDTH) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Width, FT_FLOAT, &ContainerWidth);
            Value->Double = Self->Width / ContainerWidth * 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Width, FD_FLOAT, &ContainerWidth);
            if (Self->Dimensions & DMF_FIXED_X) XCoord = Self->XCoord;
            else XCoord = Self->XCoord * 100 / ContainerWidth;
            if (Self->Dimensions & DMF_FIXED_XOFFSET) XOffset = Self->XOffset;
            else XOffset = Self->XOffset * 100 / ContainerWidth;
            Value->Double = (ContainerWidth - XOffset - XCoord) * 100 / ContainerWidth;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));

   return(ERR_Okay);
}

static ERROR SET_Width(struct Movement *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_WIDTH) | DMF_FIXED_WIDTH;
      Self->Width      = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_WIDTH) | DMF_RELATIVE_WIDTH;
      Self->Width      = Value->Double;
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:     XCoord
Synonyms: X
Short:    The horizontal position of a Movement object's monitored area.
Type:     FLOAT
Status:   Get/Set

###DESCRIPTION###
<p>If you want a Movement object to monitor a specific region of the drawable
it is connected to, set the XCoord, YCoord, Width and Height fields.  If you
would rather have the Movement object monitor the entire area of the drawable,
do not set any of the fields.</p>

<p>To set a relative/percentage based value, you must use the FD_PERCENT flag
or the value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Field: YCoord

###END###

*****************************************************************************/

static ERROR GET_X(struct Movement *Self, struct Variable *Value)
{
   OBJECTPTR Render;
   FLOAT ContainerWidth, Width;

   if (Value->Type IS FD_DOUBLE) {
      Value->Double = 0;
      if (Self->Dimensions & DMF_FIXED_X) Value->Double = Self->XCoord;
      else if (Self->Dimensions & DMF_RELATIVE_X) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Width, FT_FLOAT, &ContainerWidth);
            Value->Double = Self->XCoord * ContainerWidth / 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Width, FT_FLOAT, &ContainerWidth);
            if (Self->Dimensions & DMF_FIXED_WIDTH) Width = Self->Width;
            else Width = ContainerWidth * Self->Width / 100;
            if (Self->Dimensions & DMF_FIXED_XOFFSET) Value->Double = ContainerWidth - Width - Self->XOffset;
            else Value->Double = ContainerWidth - Width - (ContainerWidth * Self->XOffset / 100);
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_X(struct Movement *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_X) | DMF_FIXED_X;
      Self->XCoord     = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_X) | DMF_RELATIVE_X;
      Self->XCoord     = Value->Double;
   }
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XOffset
Short:  Specifies the horizontal offset of a Movement object's monitored area.
Type:   FLOAT
Status: Get/Set

###DESCRIPTION###
<p>The XOffset has a dual purpose depending on whether or not it is set in
conjunction with an X coordinate or a Width based field.</p>

<p>If set in conjunction with an X coordinate then the monitored area will be
from that X coordinate up to the width of the container, minus the value
given in the XOffset.  This means that the width of the area is dynamically
calculated in relation to the width of the container.</p>

<p>If the XOffset field is set in conjunction with a fixed or relative width
then the monitored area will start from an X coordinate calculated from the
formula &quot;XCoord = ContainerWidth - Width - XOffset&quot;.</p>

###SEE ALSO###
Field: Width, XCoord, YOffset

###END###

*****************************************************************************/

static ERROR GET_XOffset(struct Movement *Self, struct Variable *Value)
{
   OBJECTPTR Drawable;
   FLOAT ContainerWidth, Width;

   if (Value->Type IS FD_DOUBLE) {
      Value->Double = 0;
      if (Self->Dimensions & DMF_FIXED_XOFFSET) {
         Value->Double = Self->XOffset;
      }
      else if (Self->Dimensions & DMF_RELATIVE_XOFFSET) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Width, FT_FLOAT, &ContainerWidth);
            Value->Double = Self->XOffset * ContainerWidth / 100;
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH))) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Width, FT_FLOAT, &ContainerWidth);
            if (Self->Dimensions & DMF_FIXED_WIDTH) Width = Self->Width;
            else Width = ContainerWidth * Self->Width / 100;
            if (Self->Dimensions & DMF_FIXED_X) Value->Double = ContainerWidth - (Self->XCoord + Width);
            else Value->Double = ContainerWidth - ((Self->XCoord * ContainerWidth / 100) + Width);
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_XOffset(struct Movement *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_XOFFSET) | DMF_FIXED_XOFFSET;
      Self->XOffset    = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_XOFFSET) | DMF_RELATIVE_XOFFSET;
      Self->XOffset    = Value->Double;
   }
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:     YCoord
Synonyms: Y
Short:    The vertical position of a Movement object's monitored area.
Type:     FLOAT
Status:   Get/Set

###DESCRIPTION###
<p>If you want a Movement object to monitor a specific region of the drawable
it is connected to, set the XCoord, YCoord, Width and Height fields.  If you
would rather have the Movement object monitor the entire area of the drawable,
do not set any of the fields.</p>

<p>To set a relative/percentage based value, you must use the FD_PERCENT flag
or the value will be interpreted as fixed.  Negative values are permitted.</p

###SEE ALSO###
Fields: Height, XCoord, XOffset, YOffset

###END###

*****************************************************************************/

static ERROR GET_Y(struct Movement *Self, struct Variable *Value)
{
   OBJECTPTR Drawable;
   FLOAT ContainerHeight, Height;

   if (Value->Type IS FD_DOUBLE) {
      if (Self->Dimensions & DMF_FIXED_Y) {
         Value->Double = Self->YCoord;
      }
      else if (Self->Dimensions & DMF_RELATIVE_Y) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Height, FT_FLOAT, &ContainerHeight);
            Value->Double = Self->YCoord * ContainerHeight / 100;
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Height, FT_FLOAT, &ContainerHeight);
            if (Self->Dimensions & DMF_FIXED_HEIGHT) Height = Self->Height;
            else Height = ContainerHeight * Self->Height / 100;
            if (Self->Dimensions & DMF_FIXED_YOFFSET) Value->Double = ContainerHeight - Height - Self->YOffset;
            else Value->Double = ContainerHeight - Height - (ContainerHeight * Self->YOffset / 100);
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {

   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Y(struct Movement *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_Y) | DMF_FIXED_Y;
      Self->YCoord     = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_Y) | DMF_RELATIVE_Y;
      Self->YCoord     = Value->Double;
   }
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YOffset
Short:  Specifies the vertical offset of a Movement object's monitored area.
Type:   FLOAT
Status: Get/Set

###DESCRIPTION###
<p>The YOffset has a dual purpose depending on whether or not it is set in
conjunction with a Y coordinate or a Height based field.</p>

<p>If set in conjunction with a Y coordinate then the monitored area will be
from that Y coordinate up to the height of the container, minus the value
given in the YOffset.  This means that the height of the area is dynamically
calculated in relation to the height of the container.</p>

<p>If the YOffset field is set in conjunction with a fixed or relative height
then the monitored area will start from a Y coordinate calculated from the
formula &quot;YCoord = ContainerHeight - Height - YOffset&quot;.</p>

###SEE ALSO###
Field: Height, XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_YOffset(struct Movement *Self, struct Variable *Value)
{
   OBJECTPTR Drawable;
   FLOAT ContainerHeight, Height;

   if (Value->Type IS FD_DOUBLE) {
      Value->Double = 0;
      if (Self->Dimensions & DMF_FIXED_YOFFSET) {
         Value->Double = Self->YOffset;
      }
      else if (Self->Dimensions & DMF_RELATIVE_YOFFSET) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Height, FT_FLOAT, &ContainerHeight);
            Value->Double = Self->YOffset * ContainerHeight / 100;
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Height, FT_FLOAT, &ContainerHeight);
            if (Self->Dimensions & DMF_FIXED_HEIGHT) Height = Self->Height;
            else Height = ContainerHeight * Self->Height / 100;
            if (Self->Dimensions & DMF_FIXED_Y) Value->Double = ContainerHeight - (Self->YCoord + Height);
            else Value->Double = ContainerHeight - ((Self->YCoord * ContainerHeight / 100) + Height);
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));

   return(ERR_Okay);
}

static ERROR SET_YOffset(struct Movement *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_YOFFSET) | DMF_FIXED_YOFFSET;
      Self->YOffset    = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_YOFFSET) | DMF_RELATIVE_YOFFSET;
      Self->YOffset    = Value->Double;
   }
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}
