/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Keyboard
Version:   2.0
Status:    Stable
ID:        ID_KEYBOARD
Category:  I/O
Date:      March 2003
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 1996-2003.  All rights reserved.
Short:     The keyboard class is used to manage keyboard I/O.
Keywords:  keyboard, key, input, output

###DESCRIPTION###
<p>The Keyboard class is a dummy class type that is used to create objects
that represent key-based hardware devices.  This means that Keyboard objects
don't actually read hardware, they just receive input from the devices
responsible for controlling keyboard I/O.  Objects that subscribe to a
Keyboard object's data channel can then receive this input in a format that
is standardised by the Pandora Engine.</p>

<p>The most common keyboard object is the 'SystemKeyboard', which is
created during the boot process.  This particular object represents the
main PC keyboard.  If you are designing a system in which you want multiple
keyboards, you can create as many keyboard objects as you have hardware or
virtual devices to associate with them.  This can be useful for hooking up
remote keyboards over a network, or attaching extra devices that have their
own keyboard to a PC.</p>

<p>The Keyboard class supports translation methods for converting raw
key values into characters and vice versa.  Please use them rather than
translating the raw key values yourself, or you will lose vital keymap
translations that may be required on behalf of the user.</p>

###END###

*****************************************************************************/

#ifdef __snap__
/* Must include linux stuff first */
#include <sys/ioctl.h>
#include <sys/kd.h>
#include <termios.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <sched.h>
#endif

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/input/all.h>
#include <pandora/data/config.h>
#include <pandora/files/file.h>
#include <pandora/functions/strtoint.c>
#include <pandora/functions/strcopy.c>
#include <pandora/functions/writeutf8.c>
#include <pandora/input/keymaptable.h>
#include <pandora/modules/strings.h>

#define UNICODE_TABLE
#include <pandora/input/unicodetable.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_KEYBOARD 2.0

EXPORT struct ModHeader ModHeader = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_KEYBOARD, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 1996-2003.  All rights reserved.",
   "March 2003",
   "Keyboard"
};

static struct KernelBase *KernelBase = NULL;
static struct StringsBase *StringsBase = NULL;
static OBJECTPTR KeyboardClass = NULL;
static OBJECTPTR modStrings = NULL;

#ifdef __snap__

#define KEYBOARD_THREAD TRUE /* Enables threaded keyboard management */

#ifdef KEYBOARD_THREAD
#define SIZE_KEYBUFFER 30
static UBYTE glKeyboardStack[2084];
static LONG glParentProcess = 0;
static LONG glKeyPipe[2];
static LONG glKeyboardThreadID = -1;
#endif

static int KeyboardThread(void *);

static struct SNAPBase *SNAPBase;
static OBJECTPTR modSNAP = NULL;
static struct termios glTermStatus;
static LONG glKBMode = NULL;
static LONG glLEDFlags = NULL;
static LONG glKeyFlags = NULL;
static LONG glKeyState = NULL;
static UBYTE glRepeatKey[256] = { 0 };
static LONG glConsole = -1;

static UBYTE glKeyConversion[128] = {
 /* 0x00 */  NULL, K_ESCAPE, K_1, K_2, K_3, K_4, K_5, K_6, K_7, K_8, K_9, K_0, K_MINUS, K_EQUALS, K_BACKSPACE, K_TAB,
 /* 0x10 */  K_Q, K_W, K_E, K_R, K_T, K_Y, K_U, K_I, K_O, K_P, K_LSQUARE, K_RSQUARE, K_ENTER, K_LCONTROL, K_A, K_S,
 /* 0x20 */  K_D, K_F, K_G, K_H, K_J, K_K, K_L, K_SEMICOLON, K_APOSTROPHE, K_REVERSEQUOTE, K_LSHIFT, K_BACKSLASH, K_Z, K_X, K_C, K_V,
 /* 0x30 */  K_B, K_N, K_M, K_COMMA, K_DOT, K_SLASH, K_RSHIFT, K_NPMULTIPLY, K_LALT, K_SPACE, K_CAPSLOCK, K_F1, K_F2, K_F3, K_F4, K_F5,
 /* 0x40 */  K_F6, K_F7, K_F8, K_F9, K_F10, K_NUMLOCK, K_SCROLL, K_NP7, K_NP8, K_NP9, K_NPMINUS, K_NP4, K_NP5, K_NP6, K_NPPLUS, K_NP1,
 /* 0x50 */  K_NP2, K_NP3, K_NP0, K_NPDOT, K_SYSRQ, NULL, K_LESSGREATER, K_F11, K_F12, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
 /* 0x60 */  K_NPENTER, K_RCONTROL, K_NPDIVIDE, K_PRTSCR, K_RALT, K_BREAK, K_HOME, K_UP, K_PAGEUP, K_LEFT, K_RIGHT, K_END, K_DOWN, K_PAGEDOWN, K_INSERT, K_DELETE,
 /* 0x70 */  K_MACRO, K_F13, K_F14, K_HELP, K_REDO, K_F17, NULL, K_PAUSE, NULL, NULL, NULL, NULL, NULL, K_LCOMMAND, K_RCOMMAND, K_MENU
};

#endif

/*****************************************************************************
** Class definition
*/

#define LEN_LAYOUT 30

struct Keyboard {
   OBJECT_HEADER
   LONG XShift;

   PRIVATE_FIELDS
   UBYTE prvLayout[LEN_LAYOUT];    /* Name of the layout */
   UWORD prvKeymap[K_LIST_END][4]; /* Normal, +Shift, +Command, +(Shift & Command) */
   LONG  ConsoleFD;
   FLOAT RepeatRate;
   LONG  RepeatDelay;
};

static ERROR GET_Layout(struct Keyboard *, STRING *);
static ERROR GET_RepeatDelay(struct Keyboard *, LONG *);
static ERROR GET_RepeatRate(struct Keyboard *, FLOAT *);
static ERROR GET_UnicodeTable(struct Keyboard *, struct UnicodeDescriptor **);

static ERROR SET_Layout(struct Keyboard *, STRING);
static ERROR SET_RepeatDelay(struct Keyboard *, LONG *);
static ERROR SET_RepeatRate(struct Keyboard *, FLOAT *);
static ERROR SET_XShift(struct Keyboard *, LONG *);

enum {
   XSF_COMMAND=0,
   XSF_ALTGR
};

static struct FieldDef XShiftLookup[] = {
   { "Command", XSF_COMMAND },
   { "AltGr",   XSF_ALTGR },
   { NULL, NULL }
};

static struct FieldArray KeyboardFields[] = {
   { "XShift",       0, FDF_LONG|FDF_LOOKUP|FDF_RW,   (LONG)&XShiftLookup, NULL, SET_XShift },
   /*** Virtual fields ***/
   { "Layout",       0, FDF_STRING|FDF_RW, 0, GET_Layout, SET_Layout },
   { "UnicodeTable", 0, FDF_POINTER|FDF_R, 0, GET_UnicodeTable, NULL },
   { "RepeatRate",   0, FDF_FLOAT|FDF_RW,  0, GET_RepeatRate, SET_RepeatRate },
   { "RepeatDelay",  0, FDF_LONG|FDF_RW,   0, GET_RepeatDelay, SET_RepeatDelay },
   END_FIELD
};

/*****************************************************************************
** Keyboard actions.
*/

static ERROR KEYBOARD_Activate(struct Keyboard *, APTR);
static ERROR KEYBOARD_ClosingTag(struct Keyboard *, APTR);
static ERROR KEYBOARD_DataChannel(struct Keyboard *, struct acDataChannel *);
static ERROR KEYBOARD_Free(struct Keyboard *, APTR);
static ERROR KEYBOARD_Init(struct Keyboard *, APTR);
static ERROR KEYBOARD_NewObject(struct Keyboard *, APTR);
static ERROR KEYBOARD_Reset(struct Keyboard *, APTR);
static ERROR KEYBOARD_SaveToObject(struct Keyboard *, struct acSaveToObject *);
#ifdef __snap__
static ERROR KEYBOARD_Timer(struct Keyboard *, struct acTimer *);
#endif

static struct ActionArray KeyboardActions[] = {
   { AC_Activate,     KEYBOARD_Activate },
   { AC_ClosingTag,   KEYBOARD_ClosingTag },
   { AC_DataChannel,  KEYBOARD_DataChannel },
   { AC_Free,         KEYBOARD_Free },
   { AC_Init,         KEYBOARD_Init },
   { AC_NewObject,    KEYBOARD_NewObject },
   { AC_Reset,        KEYBOARD_Reset },
   { AC_SaveToObject, KEYBOARD_SaveToObject },
   #ifdef __snap__
   { AC_Timer,       KEYBOARD_Timer },
   #endif
   { NULL, NULL }
};

/*****************************************************************************
** Keyboard methods.
*/

static ERROR KEYBOARD_TranslateChars(struct Keyboard *, struct mtTranslateChars *);
static ERROR KEYBOARD_TranslateKeys(struct Keyboard *, struct mtTranslateKeys *);

static struct FunctionField argsTranslateKeys[]  = { { "Keys", ARG_PTR }, { "String", ARG_STR }, { "Size", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsTranslateChars[] = { { "String", ARG_STR }, { "Keys", ARG_PTR }, { "Size", ARG_LONG }, { NULL, NULL } };

static struct MethodArray KeyboardMethods[] = {
   { MT_TranslateKeys,  KEYBOARD_TranslateKeys,  "TranslateKeys",  argsTranslateKeys,  sizeof(struct mtTranslateKeys) },
   { MT_TranslateChars, KEYBOARD_TranslateChars, "TranslateChars", argsTranslateChars, sizeof(struct mtTranslateChars) },
   { NULL, NULL, NULL, NULL }
};

/****************************************************************************/

static ERROR LoadKeymap(struct Keyboard *);

/****************************************************************************/

static void ReadLong(OBJECTPTR File, LONG *Value)
{
   struct mtReadFileLong args;
   Action(MT_ReadFileLong, File, &args);
   if (Value) *Value = args.Result;
}

/*****************************************************************************
** Command: Init()
*/

ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
         FID_Name|TSTRING,   "strings",
         FID_Version|TFLOAT, MODVERSION_STRINGS,
         TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   #ifdef __snap__
      if (CreateObject(ID_MODULE, NULL, &modSNAP, NULL,
         FID_Name|TSTRING, "snap",
         TAGEND) IS ERR_Okay) {
         if (GetField(modSNAP, FID_ModBase, FT_POINTER, &SNAPBase) != ERR_Okay) {
            return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
         }
      }
      else return(ERR_NoSupport);

   #endif

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&KeyboardClass, NULL,
      FID_BaseClassID|TLONG, ID_KEYBOARD,
      FID_Version|TFLOAT,    VER_KEYBOARD,
      FID_Name|TSTRING,      "Keyboard",
      FID_Category|TLONG,    CCF_IO,
      FID_Fields|TPTR,       KeyboardFields,
      FID_Actions|TPTR,      KeyboardActions,
      FID_Methods|TPTR,      KeyboardMethods,
      FID_Size|TLONG,        sizeof(struct Keyboard),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   #ifdef __snap__
      if (modSNAP) { Action(AC_Free, modSNAP, NULL); modSNAP = NULL; }
   #endif

   if (modStrings)   { Action(AC_Free, modStrings, NULL);   modStrings = NULL; }
   if (KeyboardClass) { Action(AC_Free, KeyboardClass, NULL); KeyboardClass = NULL; }

   return(ERR_Okay);
}

/*****************************************************************************
** Action: Activate
*/

static ERROR KEYBOARD_Activate(struct Keyboard *Self, APTR Void)
{
#ifdef __snap__
   struct termios term;

   DPrintF("Activate:","[Keyboard] Initialising console for keyboard monitoring.");

   Self->ConsoleFD = GetResource(RES_CONSOLEFD);
   glConsole = Self->ConsoleFD;

   if (ioctl(Self->ConsoleFD, KDGKBMODE, &glKBMode) IS -1) DPrintF("@Init:","[Keyboard] ioctl(KDGKBMODE) failed.");

   /*** Set LED status ***/

   ioctl(Self->ConsoleFD, KDGETLED, &glLEDFlags);
   glLEDFlags = glLEDFlags & 0x0f;

   /*** Save the current terminal status and FD flags ***/

   if (tcgetattr(Self->ConsoleFD, &glTermStatus) IS -1) DPrintF("@Init:","[Keyboard] tcgetattr() failed.");
   glKeyState = fcntl(Self->ConsoleFD, F_GETFL);

   /* Change terminal state so that we can read the keyboard.
   **
   ** --LFLAGS--
   ** ICANON: Enables characters EOF, EOL, ERASE, KILL, STATUS.
   ** ECHO:   Print characters to the terminal.
   ** ISIG:   Generate corresponding signals on INTR, QUIT, SUSP, DSUSP
   ** --IFLAGS--
   ** BRKINT: Generate SIGINT on BREAK condition.
   ** IGNBRK: Ignore BREAK conditions.
   ** IGNPAR: Ignore framing errors and parity errors.
   */

   term = glTermStatus;

   /*** L flags ***/

   term.c_lflag &= ~(ICANON|ECHO|ISIG);

   /*** I flags ***/

   term.c_iflag &= ~(ISTRIP|IGNCR|ICRNL|INLCR|BRKINT|PARMRK|INPCK|IUCLC|IXON|IXOFF);
   term.c_iflag |= IGNBRK|IGNPAR;

   /*** Enabled/disabled key combos ***/

   term.c_cc[VMIN]  = 1;
   term.c_cc[VTIME] = 0;
   term.c_cc[VSUSP] = 0; /* Disable the suspend key combo "CTRL + Z" */

   if (tcsetattr(Self->ConsoleFD, TCSAFLUSH, &term) IS -1) DPrintF("@Init:","[Keyboard] tcsetattr() failed.");

   /*** Set raw mode so that we can read from the keyboard ***/

   if (ioctl(Self->ConsoleFD, KDSKBMODE, K_MEDIUMRAW) IS -1) DPrintF("@Init:","[Keyboard] ioctl(KDSKBMODE, K_MEDIUMRAW) failed.");

   /*** Change the FD mode to non-blocking so that we can read the keyboard without sleeping ***/

   fcntl(Self->ConsoleFD, F_SETFL, glKeyState|O_NONBLOCK);

   #ifdef KEYBOARD_THREAD
      glParentProcess = getpid();
      if (!pipe(glKeyPipe)) {
         fcntl(glKeyPipe[0], F_SETFL, fcntl(glKeyPipe[0], F_GETFL)|O_NONBLOCK);
         if ((glKeyboardThreadID = clone(&KeyboardThread, glKeyboardStack + (sizeof(glKeyboardStack)/2), CLONE_VM, NULL)) IS -1) {
            DPrintF("!!","The clone() function failed to create a keyboard monitoring process.");
         }
         else DPrintF("Init:","[Keyboard] Created keyboard monitoring thread with a process ID of #%d.", glKeyboardThreadID);
      }
   #endif

   SubscribeTimer(Self->Head.UniqueID, 20);
#endif

   return(ERR_Okay);
}

/*****************************************************************************
** Action: ClosingTag
*/

static ERROR KEYBOARD_ClosingTag(struct Keyboard *Self, APTR Void)
{
   Action(AC_Activate, Self, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  DataChannel
Short: This action can be used to send fake keyboard input.

###DESCRIPTION###
<p>Keypresses can be sent to a Keyboard object through the DataChannel action
by using the DATA_KEYPRESS data type.  All characters that are sent to a
Keyboard object through the data channel will inevitably be reported to
programs that read the Keyboard through the data channel service.</p>

<p>Please note that it is extremely important that the key qualifiers are set
correctly and that each key press is matched with a key release.  If the data
is not defined correctly, the objects receiving the data may not interpret it
as intended.</p>

###RESULT###
ERR_Okay:         The input was successfully processed.
ERR_Args:         Invalid arguments were specified.
ERR_AccessMemory: Access to the internal channel list was denied.

###END###

*****************************************************************************/

static ERROR KEYBOARD_DataChannel(struct Keyboard *Self, struct acDataChannel *Args)
{
   struct dcKeyEntry *src;
   struct acDataChannel channel;
   struct ChannelSubscription *list;
   MEMORYID ChannelListMID;
   ERROR error;
   LONG i;

   if (!Args) return(ObjectError(Self, ERH_DataChannel, ERR_Args));

   if (GetField(Self, FID_ChannelList, FT_LONG, &ChannelListMID) != ERR_Okay) return(ERR_Okay);

   if ((Args->DataType IS DATA_KEYPRESS) AND (Args->Version IS 1)) {
      struct dcKeyEntry array[Args->TotalEntries];

      /* Send all the data directly to all objects that have
      ** opened data channels with us.
      */

      channel.ObjectID     = GetUniqueID(Self);
      channel.DataType     = DATA_KEYPRESS;
      channel.Version      = 1;
      channel.Buffer       = array;
      channel.Size         = Args->TotalEntries * sizeof(struct dcKeyEntry);
      channel.TotalEntries = Args->TotalEntries;

      src = Args->Buffer;
      for (i=0; i < Args->TotalEntries; i++) {
         array[i].Flags = src[i].Flags;
         array[i].Value = src[i].Value;

         /*** Determine whether or not the keypress will translate to a printable character ***/

         array[i].Flags |= KQ_NOTPRINTABLE;
         if (array[i].Value < K_LIST_END) {

            if (array->Flags & KQ_CONTROL) continue;

            if (Self->XShift IS XSF_ALTGR) {
               if (array->Flags & (KQ_LALT|KQ_COMMAND)) continue;
            }
            else if (array->Flags & KQ_ALT) continue;

            if ((Self->XShift IS XSF_ALTGR) AND (array->Flags & KQ_RALT)) {
               if (array->Flags & KQ_SHIFT) {
                  if (Self->prvKeymap[array[i].Value][3]) array[i].Flags &= ~KQ_NOTPRINTABLE;
               }
               else {
                  if (Self->prvKeymap[array[i].Value][2]) array[i].Flags &= ~KQ_NOTPRINTABLE;
               }
            }
            else if ((Self->XShift IS XSF_COMMAND) AND (array->Flags & KQ_COMMAND)) {
               if (array->Flags & KQ_SHIFT) {
                  if (Self->prvKeymap[array[i].Value][3]) array[i].Flags &= ~KQ_NOTPRINTABLE;
               }
               else {
                  if (Self->prvKeymap[array[i].Value][2]) array[i].Flags &= ~KQ_NOTPRINTABLE;
               }
            }
            else if (array->Flags & KQ_SHIFT) {
               if (Self->prvKeymap[array[i].Value][1]) array[i].Flags &= ~KQ_NOTPRINTABLE;
            }
            else if (Self->prvKeymap[array[i].Value][0]) array[i].Flags &= ~KQ_NOTPRINTABLE;
         }
      }

      if (ChannelListMID) {
         if (AccessMemory(ChannelListMID, MEM_READ, (APTR *)&list) IS ERR_Okay) {
            i = 0;
            while (list[i].SubscriberID) {
               error = ActionMsgPort(AC_DataChannel, list[i].SubscriberID, &channel, list[i].MessagePortMID, list[i].ClassID);
               if (error IS ERR_NoMatchingObject) {
                  DPrintF("DataChannel:","[Keyboard:%d] Subscriber #%d no longer exists, removing from channel list.", Self->Head.UniqueID, list[i].SubscriberID);
                  if (UnsubscribeChannel(Self, list[i].SubscriberID) != ERR_Okay) i++;
               }
               else if (error != ERR_Okay) {
                  DPrintF("@DataChannel:","[Keyboard] Error %d on subscriber %d.", error, list[i].SubscriberID);
                  i++;
               }
               else i++;
            }
            ReleaseMemory(list);
         }
         else {
            DPrintF("@Keyboard:","Failed to gain access to data channel block #%d.", ChannelListMID);
            return(ERR_AccessMemory);
         }
      }
   }
   else if (Args->DataType IS DATA_TEXT) {
      DPrintF("@DataChanel:","[Keyboard] Text data is not yet supported.");
   }
   else DPrintF("@DataChannel:","[Keyboard] Unrecognised data type %d, version %d.", Args->DataType, Args->Version);

   return(ERR_Okay);
}

/*****************************************************************************
** Action: Free
*/

static ERROR KEYBOARD_Free(struct Keyboard *Self, APTR Void)
{
#ifdef __snap__
   LONG console;

   #ifdef KEYBOARD_THREAD
      if (glKeyboardThreadID != -1) {
         DPrintF("Free:","[Keyboard] Removing keyboard thread #%d.", glKeyboardThreadID);
         kill(glKeyboardThreadID, SIGKILL);
         glKeyboardThreadID = -1;
      }
   #endif


   if ((console = GetResource(RES_CONSOLEFD)) != -1) {
      ioctl(console, KDSKBMODE, glKBMode);
      ioctl(console, KDSETLED, glLEDFlags);
      tcsetattr(console, TCSAFLUSH, &glTermStatus); /* Return the original terminal state */
      fcntl(console, F_SETFL, glKeyState);
   }
#endif

   return(ERR_Okay);
}

/*****************************************************************************
** Action: Init
*/

static ERROR KEYBOARD_Init(struct Keyboard *Self, APTR Void)
{
   LoadKeymap(Self);
   return(ERR_Okay);
}

/*****************************************************************************
** Action: NewObject
*/

static ERROR KEYBOARD_NewObject(struct Keyboard *Self, APTR Void)
{
   struct mtReadConfigInt readint;
   struct mtReadConfigFloat readflt;
   struct mtReadConfig read;
   OBJECTPTR config;
   WORD i;

   Self->RepeatRate  = 10.9;
   Self->RepeatDelay = 250;

   /*** Set the default layout to US101 ***/

   for (i=0; "American - US101"[i]; i++) Self->prvLayout[i] = "American - US101"[i];
   Self->prvLayout[i] = 0;

   /* Attempt to get the user's preferred layout
   ** from the user:config/keyboard file.
   */

   if (CreateObject(ID_CONFIG, NF_CHILD, (OBJECTPTR *)&config, NULL,
      FID_Location|TSTRING, "user:config/keyboard",
      TAGEND) IS ERR_Okay) {

      read.Section = "KEYBOARD";
      read.Item = "Layout";
      if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
         SetField(Self, FID_Layout, FT_STRING, read.Data);
      }

      readflt.Section = "KEYBOARD";
      readflt.Item = "RepeatRate";
      if (Action(MT_ReadConfigFloat, config, &readflt) IS ERR_Okay) Self->RepeatRate = readflt.Float;

      readint.Section = "KEYBOARD";
      readint.Item = "RepeatDelay";
      if (Action(MT_ReadConfigInt, config, &readint) IS ERR_Okay) Self->RepeatDelay = readint.Integer;

      readint.Item = "XShift";
      if (Action(MT_ReadConfigInt, config, &readint) IS ERR_Okay) Self->XShift = readint.Integer;

      Action(AC_Free, config, NULL);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Reset
Short: Resets the keyboard settings.
###END###

*****************************************************************************/

static ERROR KEYBOARD_Reset(struct Keyboard *Self, APTR Void)
{
   Self->RepeatDelay = 250;
   Self->RepeatRate = 10.9;
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  SaveToObject
Short: Saves the current keyboard settings to another object.
###END###

*****************************************************************************/

static ERROR KEYBOARD_SaveToObject(struct Keyboard *Self, struct acSaveToObject *Args)
{
   struct mtWriteConfig write;
   OBJECTPTR config;
   UBYTE buffer[28];

   if (CreateObject(ID_CONFIG, NULL, &config, NULL,
      TAGEND) IS ERR_Okay) {

      write.Section = "KEYBOARD";
      write.Item = "Layout";
      write.Data = Self->prvLayout;
      Action(MT_WriteConfig, config, &write);

      write.Item = "RepeatDelay";
      write.Data = buffer;
      IntToStr(Self->RepeatDelay, buffer, sizeof(buffer));
      Action(MT_WriteConfig, config, &write);

      write.Item = "RepeatRate";
      write.Data = buffer;
      FloatToStr(Self->RepeatRate, buffer, sizeof(buffer));
      Action(MT_WriteConfig, config, &write);

      write.Item = "XShift";
      write.Data = buffer;
      IntToStr(Self->XShift, buffer, sizeof(buffer));
      Action(MT_WriteConfig, config, &write);

      ActionTags(AC_SaveToObject, config, Args->DestinationID, NULL);

      Action(AC_Free, config, NULL);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Action: Timer
*/

#ifdef __snap__

static ERROR KEYBOARD_Timer(struct Keyboard *Self, struct acTimer *Args)
{
#ifdef KEYBOARD_THREAD

   struct acDataChannel channel;
   struct dcKeyEntry entry;
   UBYTE buf[32];
   LONG i, nkeys;
   UBYTE key;

   while ((nkeys = read(glKeyPipe[0], buf, sizeof(buf))) > 0) {
      for (i=0; i < nkeys; i++) {
         key = buf[i];

         entry.Flags = NULL;
         entry.Value = NULL;

         if (key & 0x80) {
            key = glKeyConversion[key & 0x7f];
            if (!key) key = K_SPACE;

            entry.Flags |= KQ_RELEASED;
            switch (key) {
               case K_LSHIFT:   glKeyFlags &= ~KQ_LSHIFT; break;
               case K_RSHIFT:   glKeyFlags &= ~KQ_RSHIFT; break;
               case K_LCONTROL: glKeyFlags &= ~KQ_LCONTROL; break;
               case K_RCONTROL: glKeyFlags &= ~KQ_RCONTROL; break;
               case K_LALT:     glKeyFlags &= ~KQ_LALT; break;
               case K_RALT:     glKeyFlags &= ~KQ_RALT; break;
               case K_LCOMMAND: glKeyFlags &= ~KQ_LCOMMAND; break;
               case K_RCOMMAND: glKeyFlags &= ~KQ_RCOMMAND; break;
            }
            entry.Flags |= glKeyFlags;
            entry.Value = key;
            glRepeatKey[key] = 0;
         }
         else {
            key = glKeyConversion[key];
            if (!key) key = K_SPACE;

            switch (key) {
               case K_LSHIFT:   glKeyFlags |= KQ_LSHIFT; break;
               case K_RSHIFT:   glKeyFlags |= KQ_RSHIFT; break;
               case K_LCONTROL: glKeyFlags |= KQ_LCONTROL; break;
               case K_RCONTROL: glKeyFlags |= KQ_RCONTROL; break;
               case K_LALT:     glKeyFlags |= KQ_LALT; break;
               case K_RALT:     glKeyFlags |= KQ_RALT; break;
               case K_LCOMMAND: glKeyFlags |= KQ_LCOMMAND; break;
               case K_RCOMMAND: glKeyFlags |= KQ_RCOMMAND; break;

               case K_CAPSLOCK:
                  glKeyFlags ^= KQ_CAPSLOCK;
                  glLEDFlags ^= LED_CAP;
                  ioctl(Self->ConsoleFD, KDSETLED, glLEDFlags);
                  break;

               case K_NUMLOCK:
                  glKeyFlags ^= KQ_NUMLOCK;
                  glLEDFlags ^= LED_NUM;
                  ioctl(Self->ConsoleFD, KDSETLED, glLEDFlags);
                  break;

               case K_SCROLL:
                  glKeyFlags ^= KQ_SCROLLLOCK;
                  glLEDFlags ^= LED_SCR;
                  ioctl(Self->ConsoleFD, KDSETLED, glLEDFlags);
                  break;
            }

            if (glRepeatKey[key]) entry.Flags |= KQ_REPEAT;
            glRepeatKey[key]++;

            entry.Flags |= glKeyFlags|KQ_PRESSED;
            entry.Value = key;
         }

         channel.ObjectID     = NULL;
         channel.DataType     = DATA_KEYPRESS;
         channel.Version      = 1;
         channel.Buffer       = &entry;
         channel.Size         = sizeof(struct dcKeyEntry);
         channel.TotalEntries = 1;
         Action(AC_DataChannel, Self, &channel);
      }
   }

#else

   struct acDataChannel channel;
   struct dcKeyEntry entry;
   UBYTE buf[32];
   LONG i, nkeys, key;

   while ((nkeys = read(Self->ConsoleFD, buf, sizeof(buf))) > 0) {
      for (i=0; i < nkeys; i++) {
         key = buf[i];

         entry.Flags = NULL;
         entry.Value = NULL;

         if (key & 0x80) {
            key = glKeyConversion[c & 0x7f];
            if (!key) key = K_SPACE;
            entry.Flags |= KQ_RELEASED;
            switch (key) {
               case K_LSHIFT:   glKeyFlags &= ~KQ_LSHIFT; break;
               case K_RSHIFT:   glKeyFlags &= ~KQ_RSHIFT; break;
               case K_LCONTROL: glKeyFlags &= ~KQ_LCONTROL; break;
               case K_RCONTROL: glKeyFlags &= ~KQ_RCONTROL; break;
               case K_LALT:     glKeyFlags &= ~KQ_LALT; break;
               case K_RALT:     glKeyFlags &= ~KQ_RALT; break;
               case K_LCOMMAND: glKeyFlags &= ~KQ_LCOMMAND; break;
               case K_RCOMMAND: glKeyFlags &= ~KQ_RCOMMAND; break;
            }
            entry.Flags |= glKeyFlags;
            entry.Value = c;
            glRepeatKey[c] = 0;
         }
         else {
            key = glKeyConversion[key];
            if (!key) key = K_SPACE;

            switch (key) {
               case K_LSHIFT:   glKeyFlags |= KQ_LSHIFT; break;
               case K_RSHIFT:   glKeyFlags |= KQ_RSHIFT; break;
               case K_LCONTROL: glKeyFlags |= KQ_LCONTROL; break;
               case K_RCONTROL: glKeyFlags |= KQ_RCONTROL; break;
               case K_LALT:     glKeyFlags |= KQ_LALT; break;
               case K_RALT:     glKeyFlags |= KQ_RALT; break;
               case K_LCOMMAND: glKeyFlags |= KQ_LCOMMAND; break;
               case K_RCOMMAND: glKeyFlags |= KQ_RCOMMAND; break;

               case K_CAPSLOCK:
                  glKeyFlags ^= KQ_CAPSLOCK;
                  glLEDFlags ^= LED_CAP;
                  ioctl(Self->ConsoleFD, KDSETLED, glLEDFlags);
                  break;

               case K_NUMLOCK:
                  glKeyFlags ^= KQ_NUMLOCK;
                  glLEDFlags ^= LED_NUM;
                  ioctl(Self->ConsoleFD, KDSETLED, glLEDFlags);
                  break;

               case K_SCROLL:
                  glKeyFlags ^= KQ_SCROLLLOCK;
                  glLEDFlags ^= LED_SCR;
                  ioctl(Self->ConsoleFD, KDSETLED, glLEDFlags);
                  break;
            }

            if (glRepeatKey[c]) entry.Flags |= KQ_REPEAT;
            glRepeatKey[c]++;

            entry.Flags |= glKeyFlags|KQ_PRESSED;
            entry.Value = c;
         }

         if ((entry.Flags & KQ_ALT) AND (entry.Flags & KQ_CTRL)) {
            if ((entry.Value >= K_F1) AND (entry.Value <= K_F9)) {
               /*** Holding Ctrl-Alt-F1/9 switches terminals ***/
               if (ioctl(Self->ConsoleFD, VT_ACTIVATE, entry.Value - K_F1 + 1) IS -1) {
                  DPrintF("@Keyboard:","Failed to switch to terminal %d.", entry.Value - K_F1 + 1);
               }
               else DPrintF("Keyboard:","Switched to terminal %d.", entry.Value - K_F1 + 1);
            }
         }

         channel.ObjectID     = NULL;
         channel.DataType     = DATA_KEYPRESS;
         channel.Version      = 1;
         channel.Buffer       = &entry;
         channel.Size         = sizeof(struct dcKeyEntry);
         channel.TotalEntries = 1;
         Action(AC_DataChannel, Self, &channel);
      }
   }
#endif

   return(ERR_Okay);
}

/****************************************************************************/

#ifdef KEYBOARD_THREAD

#include <stdio.h>
#undef NULL
#define NULL 0

static int KeyboardThread(void *Args)
{
   struct timeval timeout;
   UBYTE buf[32];
   LONG i, nkeys, key, keyflags, result;
   fd_set fds;

   keyflags = NULL;

   while (1) {
      FD_ZERO(&fds);
      FD_SET(glConsole, &fds);

      /* Set a timeout so that we can check the existance of our parent process
      ** every one second while waiting for incoming keyboard information.  We do
      ** this due to the fact that when the parent dies, it does not automatically
      ** result in this keyboard thread being killed.
      */

      timeout.tv_sec = 1;
      timeout.tv_usec = 0;
      result = select(glConsole+1, &fds, NULL, NULL, &timeout); /* Wait for activity */

      if (glConsole IS -1) break;

      if (!result) {
         if (kill(glParentProcess, 0) IS -1) break;
         continue;
      }

      while ((nkeys = read(glConsole, buf, sizeof(buf))) > 0) {
         write(glKeyPipe[1], buf, nkeys);

         for (i=0; i < nkeys; i++) {
            key = buf[i];

            if (key & 0x80) {
               /*** Key released ***/
               key = glKeyConversion[key & 0x7f];
               if (!key) key = K_SPACE;

               switch (key) {
                  case K_LCONTROL: keyflags &= ~KQ_LCONTROL; break;
                  case K_RCONTROL: keyflags &= ~KQ_RCONTROL; break;
                  case K_LALT:     keyflags &= ~KQ_LALT; break;
                  case K_RALT:     keyflags &= ~KQ_RALT; break;
               }
            }
            else {
               key = glKeyConversion[key];
               if (!key) key = K_SPACE;

               switch (key) {
                  case K_LCONTROL: keyflags |= KQ_LCONTROL; break;
                  case K_RCONTROL: keyflags |= KQ_RCONTROL; break;
                  case K_LALT:     keyflags |= KQ_LALT; break;
                  case K_RALT:     keyflags |= KQ_RALT; break;
               }
            }

            if ((keyflags & KQ_CTRL) AND (keyflags & KQ_ALT) AND (key IS K_BACKSPACE)) {
               printf("User termination - CTRL + ALT + BACKSPACE.\n");
               kill(glParentProcess, SIGTERM);
               return(0);
            }

            if ((keyflags & KQ_CTRL) AND (keyflags & KQ_ALT) AND (key IS K_C)) {
               printf("User termination - CTRL + ALT + C.\n");
               kill(glParentProcess, SIGTERM);
               return(0);
            }
         }
      }
   }

   return(0);
}

#endif

#endif

/*****************************************************************************

###METHOD###
Name:    TranslateChars()
Short:   Translates printable characters into their equivalent key codes.
Version: 2.0

###DESCRIPTION###
<p>The TranslateChars method is used to convert UTF-8 character strings
into their equivalent raw key values.  The routine will scan the character
buffer in the String argument and write the equivalent key values to the
buffer referred to by the Keys argument.  The routine will stop when the
String buffer is terminated, or the Keys buffer runs out of space.</p>

###INPUT###
STRING/String: String of character codes to be translated.
APTR/Keys: Output buffer (struct dcKeyEntry).
LONG/Size: Size of the output buffer in bytes.

###RESULT###
ERR_Okay: The characters were successfully translated.
ERR_Args: Invalid arguments were detected.

###END###

*****************************************************************************/

static ERROR KEYBOARD_TranslateChars(struct Keyboard *Self, struct mtTranslateChars *Args)
{
   if ((!Args) OR (!Args->String) OR (!Args->Keys) OR (Args->Size < sizeof(struct dcKeyEntry))) {
      return(ObjectError(Self, MT_TranslateChars, ERR_Args));
   }





   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:    TranslateKeys()
Short:   Translates key codes into printable character values
Version: 2.0

###DESCRIPTION###
<p>The TranslateKeys method is used to convert raw key values into printable
characters.  The routine will scan the key array passed in the Keys argument
and each key value that is linked to a printable character will be written
to the String buffer that you have supplied.  The routine will stop when the
the Keys array is terminated, or the supplied buffer runs out of space.</p>

<p>Please note that this method will return unicode (UTF-8) characters for
keypresses that translate to unicode values.  For this reason the minimum
recommended buffer size is 7 bytes.  It is also worth keeping in mind that
even a single keypress can potentially return multiple characters if a string
sequence has been assigned to the key.</p>

###INPUT###
APTR/Keys: List of keys to translate (struct dcKeyEntry), terminated with a NULL entry.
STRING/String: String based output buffer.
LONG/Size: Size of the output buffer in bytes.

###RESULT###
ERR_Okay: The characters were successfully translated.
ERR_Args: Invalid arguments were detected.
ERR_EmptyString: None of the keys were translated into printable characters.

###END###

*****************************************************************************/

static ERROR KEYBOARD_TranslateKeys(struct Keyboard *Self, struct mtTranslateKeys *Args)
{
   struct dcKeyEntry *array;
   LONG pos;
   UWORD value;

   if ((!Args) OR (!Args->Keys) OR (!Args->String) OR (Args->Size < 1)) {
      return(ObjectError(Self, MT_TranslateKeys, ERR_Args));
   }

   pos = 0;
   for (array=Args->Keys; (array->Value) AND (pos < Args->Size-1); array++) {
      if (array->Value < K_LIST_END) {
         if (array->Flags & KQ_CONTROL) continue;

         if (Self->XShift IS XSF_ALTGR) {
            if (array->Flags & (KQ_LALT|KQ_COMMAND)) continue;
         }
         else if (array->Flags & KQ_ALT) continue;

         if ((Self->XShift IS XSF_ALTGR) AND (array->Flags & KQ_RALT)) {
            if (array->Flags & KQ_SHIFT) value = Self->prvKeymap[array->Value][3];
            else value = Self->prvKeymap[array->Value][2];
         }
         else if ((Self->XShift IS XSF_COMMAND) AND (array->Flags & KQ_COMMAND)) {
            if (array->Flags & KQ_SHIFT) value = Self->prvKeymap[array->Value][3];
            else value = Self->prvKeymap[array->Value][2];
         }
         else if (array->Flags & KQ_SHIFT) {
            value = Self->prvKeymap[array->Value][1];
         }
         else value = Self->prvKeymap[array->Value][0];

         if (value) {
            if ((array->Flags & KQ_CAPSLOCK) AND (value >= 'a') AND (value <= 'z')) value = (value - 'a') + 'A';
            pos += WriteUTF8(Args->String+pos, value);
         }
      }
   }

   Args->String[pos] = 0;

   if (pos < 1) return(ERR_EmptyString); /* No keys were translated */

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   RepeatDelay
Short:  Defines the delay between the holding of any key and the start of the key repeat rate.
Type:   LONG
Status: Get/Set

###DESCRIPTION###
<p>Defines the delay, in milliseconds, between the time of pressing a key and
the start of the key repeat rate.  The recommended default is 250ms.</p>

###END###

*****************************************************************************/

static ERROR GET_RepeatDelay(struct Keyboard *Self, LONG *Value)
{
   *Value = Self->RepeatDelay;
   return(ERR_Okay);
}

static ERROR SET_RepeatDelay(struct Keyboard *Self, LONG *Value)
{
   if (*Value <= 0) Self->RepeatDelay = *Value;
   else if (*Value > 1000) Self->RepeatDelay = 1000;
   else Self->RepeatDelay = *Value;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Layout
Short:  Determines the keyboard layout (defaults to US101).
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>The layout of a keyboard can be redefined by setting the Layout field.
The available keyboard layouts are determined by the available configuration
files in the &quot;system:config/keymaps/&quot; directory.  For instance,
if an &quot;italian&quot; file is available, you can set a layout of
&quot;Italian&quot;.</p>

<p>If you set a layout string that does not match any existing keymap files,
the keyboard object will revert to the default US101 format.</p>

###END###

*****************************************************************************/

static ERROR GET_Layout(struct Keyboard *Self, STRING *Value)
{
   *Value = Self->prvLayout;
   return(ERR_Okay);
}

static ERROR SET_Layout(struct Keyboard *Self, STRING Value)
{
   LONG i;

   if ((Value) AND (*Value)) {
      for (i=0; (i < LEN_LAYOUT-1) AND (Value[i]); i++) {
         Self->prvLayout[i] = Value[i];
      }
      Self->prvLayout[i] = 0;
   }

   if (Self->Head.ObjectFlags & NF_INITIALISED) {
      LoadKeymap(Self);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   RepeatRate
Short:  Defines the keyboard repeat rate for held keypresses.
Type:   LONG
Status: Get/Set

###DESCRIPTION###
<p>Defines the character repeat rate in characters per second.</p>

###END###

*****************************************************************************/

static ERROR GET_RepeatRate(struct Keyboard *Self, FLOAT *Value)
{
   *Value = Self->RepeatRate;
   return(ERR_Okay);
}

static ERROR SET_RepeatRate(struct Keyboard *Self, FLOAT *Value)
{
   if (*Value <= 0) Self->RepeatRate = *Value;
   else if (*Value > 50) Self->RepeatRate = 50;
   else Self->RepeatRate = *Value;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   UnicodeTable
Short:  Refers to a global unicode table for converting unicode values to names and vice versa.
Type:   struct UnicodeDescriptor *
Status: Get

###DESCRIPTION###
<p>This field refers to a unicode table that you can use to translate unicode
values into their standard system names.  The structure of
each entry in the table is defined as follows:</p>

<pre>
   struct UnicodeDescriptor {
      STRING Name;
      UWORD  Value;
   };
</pre>

<p>The table is sorted by unicode value (ascending) and is terminated with a
NULL entry in both fields.  You can view all of the content of the table
if you load the &quot;input/unicodetable.h&quot; include file.</p>

###END###

*****************************************************************************/

static ERROR GET_UnicodeTable(struct Keyboard *Self, struct UnicodeDescriptor **Table)
{
   *Table = glUnicodeTable;
   return(ERR_Okay);
}


/*****************************************************************************

###FIELD###
Name:   XShift
Short:  Configures the extended-shift key.
Type:   LONG
Status: Read/Write

###DESCRIPTION###

###END###

*****************************************************************************/

static ERROR SET_XShift(struct Keyboard *Self, LONG *Value)
{
   if ((*Value IS XSF_ALTGR) OR (*Value IS XSF_COMMAND)) {
      Self->XShift = *Value;
      return(ERR_Okay);
   }
   else {
      DPrintF("@SetField:","[Keyboard] Invalid XShift value of %d.", *Value);
      return(ERR_OutOfRange);
   }
}

/*****************************************************************************
** Internal: LoadKeymap()
*/

static ERROR LoadKeymap(struct Keyboard *Self)
{
   struct acRead read;
   UBYTE buffer[80], *byteswitch, tmpchar;
   UWORD *keytable;
   OBJECTPTR file;
   LONG i, j, pos, type, headlength, maxkeys, totalkeys, buffersize, keyvalue;

   pos = StrCopy("system:config/keymaps/", buffer, sizeof(buffer));
   StrCopy(Self->prvLayout, buffer+pos, sizeof(buffer)-pos);

   if (CreateObject(ID_FILE, NF_CHILD, (OBJECTPTR *)&file, NULL,
      FID_Location|TSTRING, buffer,
      FID_Flags|TLONG,      FL_READ|FL_APPROXIMATE,
      TAGEND) IS ERR_Okay) {

      read.Buffer = buffer;
      read.Length = 8;
      if ((Action(AC_Read, file, &read) IS ERR_Okay) AND (read.Length IS 8)) {
         if (StrCompare("KEYMAP10", buffer, 8, STR_MATCHCASE) IS ERR_Okay) {
            ReadLong(file, &headlength);
            ReadLong(file, &totalkeys);
            ReadLong(file, &maxkeys);

            ActionTags(AC_Seek, file, (DOUBLE)headlength, SEEK_START);

            /*** Read the key values ***/

            buffersize = totalkeys * maxkeys * sizeof(UWORD);
            if (AllocMemory(buffersize, MEM_DATA|MEM_NOCLEAR, &keytable, NULL) IS ERR_Okay) {
               read.Buffer = (APTR)keytable;
               read.Length = buffersize;
               Action(AC_Read, file, &read);

               #ifdef REVERSE_BYTEORDER
                  byteswitch = (UBYTE *)keytable;
                  for (i=0; i < buffersize-1; i+=2) {
                     tmpchar = byteswitch[i];
                     byteswitch[i] = byteswitch[i+1];
                     byteswitch[i+1] = tmpchar;
                  }
               #endif

               j = 0;
               for (i=0; i < totalkeys; i++) {
                  for (type=0; (type < 4) AND (type < maxkeys); type++) {
                     if ((keyvalue = keytable[j]) < K_LIST_END) {
                        Self->prvKeymap[keyvalue][type] = keytable[j+1+type];
                     }
                  }
                  j += maxkeys;
               }

               FreeMemory(keytable);
            }
         }
      }
      else {
         Action(AC_Free, file, NULL);
         DPrintF("!Keyboard:","Keymap file \"%s\" is empty.", buffer);
         return(ERR_GetField);
      }

      Action(AC_Free, file, NULL);
   }
   else {
      DPrintF("!Keyboard:","Failed to load keymap \"%s\".", buffer);
      return(ERR_CreateObject);
   }

   return(ERR_Okay);
}

/*****************************************************************************

               "Steep the spiral to your far abode
                In the wake of slumber, on visions I rode
                And fell like history through the chasm of ages
                Into the charged, foridden zones."

                     - In Flames

*****************************************************************************/



